''
'' VGA display 50-point Graph Overlay.  Very loosely based on the "waitvid.50x18.driver" by Marko Lukat
''...considerable help doubling from 25 points to 50 points by Marko Lukat!
''
VAR
  byte cogNum

PUB start(graphdata)
  return cogNum := cognew(@driver, graphdata)                        ' start the PASM video driver

PUB stop
  if cogNum
    cogstop(cogNum)
  cogNum~

CON
{
  We are squeezing pretty much every inch out of this COG! But it works!
  In a nutshell, the startup routine does the following:

  [sync code]                   [sync code]
  [startup code]  -->           [unrolled emitter loop]
  [free]                        [graph buffer]


  1: Setup: Initialize the video driver registers
  2: Setup: Move the "emitter unroll" code to a calculated destination ("pinning the tail on the donkey" with the "emitter tail" code)
  3: Setup: Run the moved "emitter unroll code".  (Can't run it in place: it'll overwrite itself!)
  4: Unroll: Duplicate the "emitter" loop 49 times from its location out.  Overwrite SETUP...all the way to a perfect match with "emitter tail"
  5: Unroll: Jump to VSync
  6: VSync: Read graph buffer into COG RAM, overwrite "emitter unroll" code.  (If the VSync code is too long, this can overwrite it [will FIT error at compile])
  7: Live happily ever after 

}
  
DAT           org     0                       ' video driver

              'NOT giving options for "configurable Y position, configurable X-range", etc.  Hardcode for one purpose.  That greatly simplifies the init routine.
              
              'NOTE: Setup is below, as it only runs once, after which we overwrite it with the unrolled emitter loop
driver        jmp #setup        ' setup -> move unroller loop -> run unroller loop -> return to VSYNC

              '==================================
              'Constants all up here to save one instruction (why JMP around the constants if you don't have to?)
vsync_pin     long %01 << (pgrp * 8)            'used for syncing to video driver 
hsync_pin     long %10 << (pgrp * 8)            'used for syncing to video driver

mask          long    vpin << (pgrp * 8) '### | %11 << (sgrp * 8) ###...removed VSync/HSync pins.  We're only trying to sync to the text screen 
dst1          long      1 << 9                  ' dst +/-= 1

mask_ch0      long $0000FE00
mask_ch1      long $00FE0000
mask_ch2      long $FE000000

grid_test     long $00001F00    ' mask for "ybase", the Y count, so the grid is not relative to the graph position.  ALSO not testing rather superfluous ycount so there's a gridline across the bottom of the graph!

ybaseStart    long $FFFFFF00    ' start pattern for graph Y count...
ybaseOffset   long $01010100    ' for add/subtract

'Can't use RES, for variables as it doesn't use any memory!
''NOTE:NOTE: Only three variables?  Fixed in CON block at bottom to DIRB/OUTB/FRQB                
'temp          long       0      'res 1
'ctrvar        long       0                       ' required to count the start Y position for the graph
'ybase         long       0                       ' calculated Y, plus the color adjustment...

' horizontal timing 800(800) 40(40) 128(128) 88(88)
'   vertical timing 600(600)  1(1)    4(4)   23(23)
'=================================================================================================
vsync         
              ''From Marko Lukat...copy the graph buffer to internal COG RAM.  That's 50 longs...
              mov ctrvar, #50                   'CONST: #of blocks across...
              mov temp, par                     ' Set the source address of the update loop
              movd updLoop, #MEM_graphbuf       ' Set the destination of the RDWORD instruction.
              movd updShft, #MEM_graphbuf       ' Set the destination of the SHIFT instruction.

              mov dira, #0                      ' Disable output--this prevents "noise" from slipping across the screen.  Can't be directly on VSync, or we chop the last grid section off!
              
updLoop       rdlong 0, temp                    ' Read a LONG, as we have 3 axis to work with.
updShft       shl 0, #8                         ' Boot it up to the high word for the emit loop.  LOW BYTE HAS TO BE CLEAR
              
              add updLoop, dst1                 ' Increment destination pointer (modify the instruction)
              add updShft, dst1                 ' Increment shift destination pointer (modify the instruction)
              add temp, #4                      ' Increment source...doing WORDs now.
              djnz ctrvar, #updLoop             ' Loop until the buffer is updated        

              ' We don't need to do the video sync/drive portion of anything.
              'Just overlaying on top of the already-generated signal.  Wait for VSync...
              waitpeq   vsync_pin, vsync_pin    ' Wait for the VSync pin to go HIGH
              waitpne   vsync_pin, vsync_pin    ' Wait for the VSync pin to go LOW
      '        mov       ctrvar, #0 'OPTIONAL--DJNZ LEAVES IT ZERO!!             ' starting at the top of the screen.
              mov       ybase, ybaseStart       ' initialize the line counter/compare register...

              ' Synchronize to the HSync signal, count lines until we're in the "graph range"
line          waitpeq   hsync_pin, hsync_pin    ' Wait for the HSync pin to go HIGH
              waitpne   hsync_pin, hsync_pin    ' Wait for the HSync pin to go LOW
              add       ctrvar, #1

              cmp       ctrvar, #graph_start    wc
        if_c  jmp       #line
              '------------------------------------------------------------------------------------------------
              'Woohoo--it synchronizes reliably, even if the "error" is +/- 8 pixels on the screen.  We can live with that.
              waitvid   zero, #0                'NOTE: The DEST here HAS to point to a LONG which has an LSByte of $00, or we get pink grid lines
              ' Write a grid by modifying the LOW byte of "ybase" constant...
              test      ybase, grid_test     wz
              muxz      yBase, #color_grid      'CONST: Output color for graph lines
              '---------------------------------
              mov       temp, ybase             ' scanline * $01010000 + %%1
              mov       dira, mask              ' drive outputs.  NOTE: Does not drive VSYNC/HSYNC pins.
              
              'Need to wait out past the blanking portion...
              '78-82 is rock-solid on the "bad" screen, with only axis 1.
              '86-96 is blank
              '97 has extremely "curly" shadow of graph
              '98-100 works, but over a few pixels too far; that's the fault of the WAITVID, and we can't do much about it.
              '101 is "curly" 

              'Add/subtract by 32 for HBlanking offsets.              
              '---------------------------------
              mov     cnt, #66                  ' 5 is minimal offset to avoid full range delay.  100 is good...68 is good...
              add     cnt, cnt                  ' current time
zero          waitcnt cnt, #0                   ' run-through

              'TEMP:
              '.byte[3] = 3rd axis
              '.byte[2] = 2nd axis
              '.byte[1] = 1st axis
              '.byte[0] = calculated color for this graph block.  Note that axis can "merge" together...
              '---------------------------------------
copy_start    sub     temp, MEM_graphbuf        ' mask all 3 output values at once.  NOTE: Has to have initial pointer to the graph buffer for the unrolling
              'NOTE: XOR doesn't have bytelane issues...but it doesn't give pixel-resolution, either.  SUB gives pixel resolution, but... 
               
              test    temp, mask_ch0 wz         ' $0000FE00
        if_e  or      temp, #color_0            'CONST: Output color for axis 1
              test    temp, mask_ch1 wz         ' $00FE0000
        if_e  or      temp, #color_1            'CONST: Output color for axis 2
              test    temp, mask_ch2 wz         ' $FE000000
        if_e  or      temp, #color_2            'CONST: Output color for axis 3

              mov     temp, ybase               ' WHOP, reinit temp
'              add     copy_start, #1            ' next data entry
'              djnz    xBlocks, #copy_start       ' next pixel

''NOTE!  The "tail" code that loops back to VSYNC is copied to the end of COG RAM by the "unroll loop move" at Setup.  Convoluted, for sure!

'=================================================================================================
'WARNING:WARNING:WARNING: Code below this line WILL get overwritten by the unrolled emitter loop
'
'  IS THERE ENOUGH MEMORY???  496 LONGs available in the COG...
'       405 LONGs for unrolled emitter loop (to get 3 axis) = 8 * 50, plus 5 for the end loop code
'        50 LONGs for the axis data buffer
'...41 LONGs left.
'
'  Graph buffer is at (496 - 50) = hardcoded at COG location #446

setup         ' Synchronize the startup of the video PLL to the VSync pin from the tile driver.
              ' About as effective at synchronization as a much more complicated CNT sync...and still has the one-block-off random error.
              waitpne   vsync_pin, vsync_pin    ' Wait for the VSync pin to go LOW
              waitpeq   vsync_pin, vsync_pin    ' Wait for the VSync pin to go HIGH
              waitpne   vsync_pin, vsync_pin    ' Wait for the VSync pin to go LOW
              '--------------------------------
              'Initialize the video hardware peripherals...        
              movi    ctrb, #%0_11111_000     ' LOGIC always (loader support)                         '###?????  VCO+128 (lower three), counter mode "ALWAYS", no output pin (upper five)
               
              movi    ctra, #%0_00001_110     ' PLL, VCO/2
              movi    frqa, #%0001_00000      ' 5MHz * 16 / 2 = 40MHz = pixel clock
              movs    frqa, #res_x/4          ' |
              movs    frqa, #0                ' insert res_x into phsa
               
              mov     vscl, hvis              ' 16/16
              mov     vcfg, vcfg_norm         ' VGA, 4 colour mode            ### was vcfg_sync
              ''NOTE: Doesn't make any difference if the "hub sync" code is present or not.  Sync issues remain.

              

              'Copy the "unroll emitter" loop to the end of COG RAM.  You know we're getting REAL tight!
              mov ctrvar, #(unrollJMP - unrollStart) + (unrollEnd - unrollORG)
moveunroll    mov copy_start + (8 * 50), unrollStart                         'copy_start + (8 * 50) = "emitStart"
              add moveunroll, unrollADD_abs   'update pointers...have to use the absolute LONG, not the "timbuktu relative"
              djnz ctrvar, #moveunroll

              jmp #unrollORG                    'The ORG directive means that we can label-jump to it...ONLY after we move it!
               '============================================================================================
' initialised data and/or presets.  Separate from the lower group, as these can be overwritten--and we need every LONG we can get!
hvis          long      16 << 12 | 16            '   16/16, single pixel stretched over whole time.  Note that the text driver is 1 << 12
vcfg_norm     long    %0_01_1_00_000 << 23 | pgrp << 9 | vpin
unrollADD_abs long (1 << 9) | 1                 ' src+dst +/- 1.  FOR SETUP "moveunroll"

              'This following code is moved by the above lines high out in COG RAM, just below the 50 LONG graph buffer.
              'The code from "unrollJMP" and onwards can be safely overwritten by the graph buffer after it's served its place.
              '...and the unroll code will overwrite from "SETUP" all the way out.  Difficult!
              '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
unrollStart   '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
              'How many rows to display?  Note that we can reuse the YBASE variable to also tell us when we're done!
              cmp       ybase, #color_grid  wz  ' Is yBase 0?  Can't use WZ on the SUB, as then we chop line 0 off.  And can't compare for 0 here, as the last line has the grid.  Sheesh!
              sub       ybase, ybaseOffset      ' decrement the "compare long", as that's how we compare the graph points for realtime display
        if_nz jmp       #line                   ' Loop until all lines done.
              'Done.  Wait for the next frame.
              jmp       #vsync                  ' next frame
              '----------------------------------
              'THE ABOVE CODE IS NOT RELATED TO THE UNROLL LOOP, EXCEPT BY LOCATION!
              'GRAPH BUFFER MUST BE BEYOND THIS, OR IT WILL CRASH.
unrollJMP
              ORG (copy_start + (8 * 50)) + (unrollJMP - unrollStart)           'FOLLOWING CODE LABELS ARE RELATIVE TO THIS!
unrollORG     fit MEM_graphbuf                  'error if above "tail" overlaps graph buffer
                            
             'Unroll the emitter loop the whole way out... 
              mov ybase, #49                    '#of iterations.  1 less, 'cause we don't overwrite the last one.
oLoop         movs iLoop, #copy_start           'Reset the copy SRC.  NOTE!!  HAS TO BE ONE ISN AWAY FROM ILOOP, OR FIRST COPY IS FROM $000!! 
              mov ctrvar, #8                    '#of LONGs to copy.  CONST 'cause it's WHOP and a fixed number
              add copy_start, #1                ' modify the source value from the XOR to the next buffer address.  Has to be before so the 1st copy (=2nd block) is correct 
              
              'Inner loop: copy 8 LONGs at a time. 
iLoop         mov copy_start + 8, 0+0           'Hardcode to 
              add iLoop, unrollADD_rel          'NEXT
              djnz ctrvar, #iLoop 
              djnz ybase, #oLoop

              'Not done yet--fix "copy_start" to point to the first byte in the array...
              movs copy_start, #MEM_graphbuf 
              jmp #vsync                        'ABS pointer...start the graph!
unrollADD_rel long (1 << 9) | 1                 ' src+dst +/- 1.  NOTE: ONLY USED BY UNROLLER LOOP.
unrollEnd     '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
              '!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

CON
  MEM_graphbuf = 496 - 50
  ctrvar = $1F5 '=OUTB.  Re-using the register...
  temp = $1F7   '=DIRB.
  ybase = $1FB  '=FRQB.  Only using CTRA for video synchronization.


  
  vpin    = $0FC                                ' pin group mask
  pgrp    = 1                                   ' pin group
  
  res_x   = 800                                 ' |
  res_y   = 600                                 ' |
  res_m   = 4                                   ' UI support

  graph_start = 343             'graph height is hardcoded at 256 pixels.
  'colors %RRGG BB00
  color_Grid = $54
  color_0 = $C0
  color_1 = $30
  color_2 = $0C